from django.views.generic.base import TemplateResponseMixin, View
from django.http import HttpResponseRedirect
from django.forms.formsets import formset_factory
from django.forms.models import modelformset_factory, inlineformset_factory
from django.views.generic.detail import SingleObjectMixin, SingleObjectTemplateResponseMixin
from django.views.generic.list import MultipleObjectMixin, MultipleObjectTemplateResponseMixin
from django.forms.models import BaseInlineFormSet


class BaseFormSetMixin(object):
    """
    Base class for constructing a FormSet within a view
    """
    
    initial = []
    form_class = None
    formset_class = None
    success_url = None
    extra = 2
    max_num = None
    can_order = False
    can_delete = False
    
    def construct_formset(self):
        return self.get_formset()(initial=self.get_initial(), **self.get_formset_kwargs())

    def get_initial(self):
        return self.initial

    def get_formset_class(self):
        return self.formset_class

    def get_form_class(self):
        return self.form_class

    def get_formset(self):
        return formset_factory(self.get_form_class(), **self.get_factory_kwargs())
    
    def get_formset_kwargs(self):
        kwargs = {}
        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return kwargs
    
    def get_factory_kwargs(self):
        kwargs = {
            'extra': self.extra,
            'max_num': self.max_num,
            'can_order': self.can_order,
            'can_delete': self.can_delete,
        }
        
        if self.get_formset_class():
            kwargs['formset'] = self.get_formset_class()
        
        return kwargs    
    

class FormSetMixin(BaseFormSetMixin):
    def get_context_data(self, **kwargs):
        return kwargs

    def get_success_url(self):
        if self.success_url:
            url = self.success_url
        else:
            # Default to returning to the same page
            url = self.request.get_full_path() 
        return url

    def formset_valid(self, formset):
        return HttpResponseRedirect(self.get_success_url())

    def formset_invalid(self, formset):
        return self.render_to_response(self.get_context_data(formset=formset))


class ModelFormSetMixin(FormSetMixin, MultipleObjectMixin):
    exclude = None
    fields = None
    formfield_callback = None

    def get_context_data(self, **kwargs):
        context = kwargs

        if self.object_list:
            context['object_list'] = self.object_list
            context_object_name = self.get_context_object_name(self.get_queryset())
            if context_object_name:
                context[context_object_name] = self.object_list
        return context

    def construct_formset(self):
        return self.get_formset()(queryset=self.get_queryset(), **self.get_formset_kwargs())

    def get_factory_kwargs(self):
        kwargs = super(ModelFormSetMixin, self).get_factory_kwargs()
        kwargs.update({
            'exclude': self.exclude,
            'fields': self.fields,
            'formfield_callback': self.formfield_callback,
        })
        if self.get_form_class():
            kwargs['form'] = self.get_form_class()
        if self.get_formset_class():
            kwargs['formset'] = self.get_formset_class()
        return kwargs
    
    def get_formset(self):
        return modelformset_factory(self.model, **self.get_factory_kwargs())
    
    def formset_valid(self, formset):
        self.object_list = formset.save()
        return super(ModelFormSetMixin, self).formset_valid(formset) 
    

class BaseInlineFormSetMixin(BaseFormSetMixin):
    model = None
    inline_model = None
    fk_name = None    
    formset_class = BaseInlineFormSet
    exclude = None
    fields = None
    formfield_callback = None
    can_delete = True

    def get_context_data(self, **kwargs):
        context = kwargs

        if self.object:
            context['object'] = self.object
            context_object_name = self.get_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        return context

    def construct_formset(self):
        return self.get_formset()(instance=self.object, **self.get_formset_kwargs())    
    
    def get_inline_model(self):
        return self.inline_model
    
    def get_factory_kwargs(self):
        kwargs = super(BaseInlineFormSetMixin, self).get_factory_kwargs()
        kwargs.update({
            'exclude': self.exclude,
            'fields': self.fields,
            'formfield_callback': self.formfield_callback,
            'fk_name': self.fk_name,
        })
        if self.get_form_class():
            kwargs['form'] = self.get_form_class()
        if self.get_formset_class():
            kwargs['formset'] = self.get_formset_class()
        return kwargs
    
    def get_formset(self):
        return inlineformset_factory(self.model, self.get_inline_model(), **self.get_factory_kwargs())
    

class InlineFormSetMixin(BaseInlineFormSetMixin, FormSetMixin, SingleObjectMixin):
    def formset_valid(self, formset):
        self.object_list = formset.save()
        return super(BaseInlineFormSetMixin, self).formset_valid(formset)   


class ProcessFormSetView(View):
    """
    A mixin that processes a fomset on POST.
    """
    def get(self, request, *args, **kwargs):
        formset = self.construct_formset()
        return self.render_to_response(self.get_context_data(formset=formset))

    def post(self, request, *args, **kwargs):
        formset = self.construct_formset()
        if formset.is_valid():
            return self.formset_valid(formset)
        else:
            return self.formset_invalid(formset)

    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)


class BaseFormSetView(FormSetMixin, ProcessFormSetView):
    """
    A base view for displaying a formset
    """


class FormSetView(TemplateResponseMixin, BaseFormSetView):
    """
    A view for displaying a formset, and rendering a template response
    """


class BaseModelFormSetView(ModelFormSetMixin, ProcessFormSetView):
    """
    A base view for displaying a modelformset
    """
    def get(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        return super(BaseModelFormSetView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        return super(BaseModelFormSetView, self).post(request, *args, **kwargs)


class ModelFormSetView(MultipleObjectTemplateResponseMixin, BaseModelFormSetView):
    """
    A view for displaying a modelformset, and rendering a template response
    """


class BaseInlineFormSetView(InlineFormSetMixin, ProcessFormSetView):
    """
    A base view for displaying a modelformset for a queryset belonging to a parent model
    """
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseInlineFormSetView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseInlineFormSetView, self).post(request, *args, **kwargs)    


class InlineFormSetView(SingleObjectTemplateResponseMixin, BaseInlineFormSetView):
    """
    A view for displaying a modelformset for a queryset belonging to a parent model
    """



from django.views.generic.base import TemplateResponseMixin, View
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect, Http404
from django.forms.formsets import formset_factory
from django.forms.models import modelformset_factory
from django.forms import models as model_forms
from django.forms import ValidationError

class FormProvider(object):   
    def __init__(self, form_class, context_suffix, init_args={}):
        self.form_class = form_class
        self.context_suffix = context_suffix
        self.init_args = init_args
        
    def get_context_suffix(self):
        return self.context_suffix
    
    def get_form(self, caller, prefix):
        kwargs = {}
        
        for k,v in self.init_args.iteritems():
            method_name = v % prefix
            try:
                kwargs[k] = getattr(caller, method_name)()
            except AttributeError:
                
                msg = '%s must implement method "%s" ' % (caller.__class__.__name__, method_name)
                raise ImproperlyConfigured(msg)
        kwargs.update(caller.get_form_kwargs(prefix))
        
        try:
            form = self.form_class(prefix=prefix, **kwargs)
        except ValidationError, e:
            # This is nasty.  Basically a formset will throw a validation error on instantiation
            # if the management form is missing, but we expect it to be empty if it wasn't one
            # of the POSTed forms, so we have to catch the error and deal with it later.
            form = e
            form.prefix = prefix
        return form


class MultiFormMixin(object):
    """
    Handle multiple forms in a single view
    """
    forms = {}
    initial = {}
    success_url = None
    groups = None
    
    @staticmethod
    def form(form):
        return FormProvider(form_class=form, context_suffix='form')
    
    @staticmethod
    def modelform(model, form=None, **kwargs):
        if not form:
            form = model_forms.modelform_factory(model, **kwargs)
        return FormProvider(form_class=form, context_suffix='form' , init_args={'instance': 'get_%s_instance'})

    @staticmethod
    def formset(form, **kwargs):
        generated_formset = formset_factory(form, **kwargs)
        return FormProvider(form_class=generated_formset, context_suffix='formset')
    
    @staticmethod
    def modelformset(model, **kwargs):
        generated_formset = modelformset_factory(model, **kwargs)
        return FormProvider(form_class=generated_formset, context_suffix='formset', init_args={'queryset': 'get_%s_queryset'})
    
    def get_initial(self,prefix):
        handler = 'get_initial_%s' % prefix
        if hasattr(self, handler):
            return getattr(self, handler)()
        return self.initial.get(prefix,{})
    
    def get_form_definitions(self):
        return self.forms
    
    def get_groups(self):
        if not self.groups:
            return { 'all': self.forms }
        return self.groups

    def construct_forms(self):
        forms = {}
        definitions = self.get_form_definitions()
        
        for prefix, provider in definitions.iteritems():
            context_name = '%s_%s' % (prefix, provider.get_context_suffix())
            forms[context_name] = provider.get_form(self, prefix)
        return forms        
    
    def get_form_kwargs(self,prefix):
        kwargs = {'initial': self.get_initial(prefix)}
        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return kwargs

    def get_context_data(self, **kwargs):
        return kwargs
            
    def get_success_url(self):
        if self.success_url:
            url = self.success_url
        else:
            raise ImproperlyConfigured(
                "No URL to redirect to. Provide a success_url.")
        return url      
            
    def forms_valid(self):
        return HttpResponseRedirect(self.get_success_url())
    
    def forms_invalid(self, forms):
        return self.render_to_response(self.get_context_data(**forms))
    

class ProcessMultiFormView(View):
    """
    The equivalent of Django's ProcessFormView but for MultiForms
    """
    def get(self, request, *args, **kwargs):
        forms = self.construct_forms()
        return self.render_to_response(self.get_context_data(**forms))
    
    def post(self, request, *args, **kwargs):
        forms = self.construct_forms()
        forms_dict = dict([(x.prefix, x) for x in forms.values()])
        
        valid = True
        valid_forms = {}
        invalid_forms = {}
        
        posted_prefixes = []
        
        # First we detect which prefixes were POSTed
        for prefix in self.get_form_definitions().keys():
            for field in self.request.POST:
                if field.startswith(prefix):
                    posted_prefixes.append(prefix)
                    break

        # Now we iterated over the groups until we find one that matches the POSTed prefixes
        for label, prefixes in self.get_groups().iteritems():
            if label == 'all' or list(prefixes) == posted_prefixes:
                # We've found the group, now check if all its forms are valid
                for prefix in prefixes:
                    form = forms_dict[prefix]

                    # FormSets force us to do this...
                    if isinstance(form, ValidationError):
                        raise form
                    
                    if form.is_valid():
                        valid_forms[prefix] = form
                    else:
                        valid = False
                        invalid_forms[prefix] = form
                if valid:
                    handler = 'valid_%s' % label
                    if hasattr(self, handler):
                        getattr(self, handler)(valid_forms)
                    return self.forms_valid()
                else:
                    handler = 'invalid_%s' % label
                    if hasattr(self, handler):
                        getattr(self, handler)(invalid_forms)                    
                    return self.forms_invalid(forms)
                break
            
        # If we got here, it means we couldn't find a matching group for the POST data
        raise Http404()
    
    def put(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)


class BaseMultiFormView(MultiFormMixin, ProcessMultiFormView):
    pass


class MultiFormView(TemplateResponseMixin, BaseMultiFormView):
    pass



from django.views.generic.dates import BaseMonthArchiveView
from django.views.generic.list import MultipleObjectTemplateResponseMixin
from django.core.exceptions import ImproperlyConfigured
from calendar import Calendar
from collections import defaultdict
import datetime


class BaseCalendarMonthArchiveView(BaseMonthArchiveView):
    first_of_week = 0  # 0 = Monday, 6 = Sunday
    paginate_by = None

    def get_paginate_by(self, queryset):
        if self.paginate_by is not None:
            raise ImproperlyConfigured(u"'%s' cannot be paginated, it is a calendar view" % self.__class__.__name__)
        return None
        
    def get_first_of_week(self):
        return self.first_of_week
    
    def get_context_data(self, **kwargs):
        data = super(BaseCalendarMonthArchiveView, self).get_context_data(**kwargs)
        date = data['date_list'][0]
        
        cal = Calendar(self.get_first_of_week())

        month_calendar = []
        now = datetime.datetime.utcnow()
        
        date_lists = defaultdict(list)
        
        for obj in data['object_list']:
            obj_date = getattr(obj, self.get_date_field())            
            try:
                obj_date = obj_date.date()
            except AttributeError:
                # It's a date rather than datetime, so we use it as is
                pass                    
            date_lists[obj_date].append(obj)

        for week in cal.monthdatescalendar(date.year, date.month):
            week_calendar = []
            for day in week:
                week_calendar.append({
                    'day': day,
                    'object_list': date_lists[day],
                    'today': day == now.date(),
                    'is_current_month': day.month == date.month,
                })
            month_calendar.append(week_calendar)
            
        data['calendar'] = month_calendar

        return data
    

class CalendarMonthArchiveView(MultipleObjectTemplateResponseMixin, BaseCalendarMonthArchiveView):
    template_name_suffix = '_calendar_month'



from django.views.generic.edit import FormView, ModelFormMixin
from django.views.generic.detail import SingleObjectTemplateResponseMixin
from extra_views.formsets import BaseInlineFormSetMixin
from django.http import HttpResponseRedirect
from django.forms.formsets import all_valid 


class InlineFormSet(BaseInlineFormSetMixin):

    def __init__(self, parent_model, request, instance):
        self.inline_model = self.model        
        self.model = parent_model
        self.request = request
        self.object = instance


class ModelFormWithInlinesMixin(ModelFormMixin):
    def forms_valid(self, form, inlines):
        self.object = form.save()
        for formset in inlines:
            formset.save()
        return HttpResponseRedirect(self.get_success_url())
    
    def forms_invalid(self, form, inlines):
        return self.render_to_response(self.get_context_data(form=form, inlines=inlines))
    
    def construct_inlines(self):
        inline_formsets = []
        for inline_class in self.inlines:
            inline_instance = inline_class(self.model, self.request, self.object)
            inline_formset = inline_instance.construct_formset()
            inline_formsets.append(inline_formset)
        return inline_formsets


class ProcessFormWithInlinesView(FormView):
    def get(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        inlines = self.construct_inlines()
        return self.render_to_response(self.get_context_data(form=form, inlines=inlines))

    def post(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        form = self.get_form(form_class)

        if form.is_valid():
            self.object = form.save(commit=False)
            form_validated = True
        else:
            form_validated = False

        inlines = self.construct_inlines()

        if all_valid(inlines) and form_validated:
            return self.forms_valid(form, inlines)
        return self.forms_invalid(form, inlines)

    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)


class BaseCreateWithInlinesView(ModelFormWithInlinesMixin, ProcessFormWithInlinesView):
    def get(self, request, *args, **kwargs):
        self.object = None
        return super(BaseCreateWithInlinesView, self).get(request, *args, **kwargs)
    
    def post(self, request, *args, **kwargs):
        self.object = None
        return super(BaseCreateWithInlinesView, self).post(request, *args, **kwargs)


class CreateWithInlinesView(SingleObjectTemplateResponseMixin, BaseCreateWithInlinesView):
    template_name_suffix = '_form'


class BaseUpdateWithInlinesView(ModelFormWithInlinesMixin, ProcessFormWithInlinesView):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseUpdateWithInlinesView, self).get(request, *args, **kwargs)
    
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseUpdateWithInlinesView, self).post(request, *args, **kwargs)


class UpdateWithInlinesView(SingleObjectTemplateResponseMixin, BaseUpdateWithInlinesView):
    template_name_suffix = '_form'


class NamedFormsetsMixin(object):
    inlines_names = []

    def get_context_data(self, **kwargs):
        ctx = super(NamedFormsetsMixin, self).get_context_data(**kwargs)
        if not self.inlines_names:
            return ctx
        # We have formset or inlines in context, but never both
        ctx.update(zip(self.inlines_names, kwargs.get('inlines',[])))
        if 'formset' in kwargs:
            ctx[self.inlines_names[0]] = kwargs['formset']
        return ctx


import functools
from django.core.exceptions import ImproperlyConfigured

class SortHelper(object):
    def __init__(self, request, sort_fields_aliases, sort_param_name, sort_type_param_name):
        self.initial_params = request.GET.copy()
        self.sort_fields = dict(sort_fields_aliases)
        self.inv_sort_fields = dict((v, k) for k, v in sort_fields_aliases)
        self.initial_sort = self.inv_sort_fields.get(self.initial_params.get(sort_param_name), None)
        self.initial_sort_type = self.initial_params.get(sort_type_param_name, 'asc')
        self.sort_param_name = sort_param_name
        self.sort_type_param_name = sort_type_param_name

        for field, alias in self.sort_fields.items():
            setattr(self, 'get_sort_query_by_%s' % alias, functools.partial(self.get_params_for_field, field))
            setattr(self, 'get_sort_query_by_%s_asc' % alias, functools.partial(self.get_params_for_field, field, 'asc'))
            setattr(self, 'get_sort_query_by_%s_desc' % alias, functools.partial(self.get_params_for_field, field, 'desc'))
            setattr(self, 'is_sorted_by_%s' % alias, functools.partial(self.is_sorted_by, field))

    def is_sorted_by(self, field_name):
        return field_name == self.initial_sort and self.initial_sort_type or False

    def get_params_for_field(self, field_name, sort_type=None):
        """
        If sort_type is None - inverse current sort for field, if no sorted - use asc
        """
        if not sort_type:
            if self.initial_sort == field_name:
                sort_type = 'desc' if self.initial_sort_type == 'asc' else 'asc'
            else:
                sort_type = 'asc'
        self.initial_params[self.sort_param_name] = self.sort_fields[field_name]
        self.initial_params[self.sort_type_param_name] = sort_type
        return '?%s' % self.initial_params.urlencode()

    def get_sort(self):
        if not self.initial_sort:
            return None
        sort = '%s' % self.initial_sort
        if self.initial_sort_type == 'desc':
            sort = '-%s' % sort
        return sort


class SortableListMixin(object):
    """
    You can provide either sort_fields as a plain list like ['id', 'some', 'foo__bar', ...]
    or, if you want to hide original field names you can provide list of tuples with aliace that will be used:
    [('id', 'by_id'), ('some', 'show_this'), ('foo__bar', 'bar')]

    If sort_param_name exists in query but sort_type_param_name is omitted queryset will be sorted as 'asc'
    """
    sort_fields = []
    sort_fields_aliases = []
    sort_param_name = 'o'
    sort_type_param_name = 'ot'

    def get_sort_fields(self):
        if self.sort_fields:
            return zip(self.sort_fields, self.sort_fields)
        return self.sort_fields_aliases

    def get_sort_helper(self):
        self.sort_helper = SortHelper(self.request, self.get_sort_fields(), self.sort_param_name, self.sort_type_param_name)
        return self.sort_helper

    def _sort_queryset(self, queryset):
        helper = self.get_sort_helper()
        sort = self.sort_helper.get_sort()
        if sort:
            queryset = queryset.order_by(sort)
        return queryset

    def get_queryset(self):
        qs = super(SortableListMixin, self).get_queryset()
        if self.sort_fields and self.sort_fields_aliases:
            raise ImproperlyConfigured('You should provide sort_fields or sort_fields_aliaces but not both')
        return self._sort_queryset(qs)

    def get_context_data(self, **kwargs):
        ctx = super(SortableListMixin, self).get_context_data(**kwargs)
        if hasattr(self, 'sort_helper'):
            ctx['sort_helper'] = self.sort_helper
        return ctx


from django.contrib.contenttypes.generic import generic_inlineformset_factory, BaseGenericInlineFormSet
from extra_views.formsets import BaseInlineFormSetMixin, InlineFormSetMixin, BaseInlineFormSetView, InlineFormSetView

class BaseGenericInlineFormSetMixin(BaseInlineFormSetMixin):
    ct_field = "content_type"
    ct_fk_field = "object_id"
    formset_class = BaseGenericInlineFormSet
    
    def get_factory_kwargs(self):
        kwargs = super(BaseGenericInlineFormSetMixin, self).get_factory_kwargs()
        del kwargs['fk_name']
        kwargs.update({
            "ct_field": self.ct_field,
            "fk_field": self.ct_fk_field,
        })
        return kwargs
    
    def get_formset(self):
        result = generic_inlineformset_factory(self.inline_model, **self.get_factory_kwargs())
        return result


class GenericInlineFormSet(BaseGenericInlineFormSetMixin):
    def __init__(self, parent_model, request, instance):
        self.inline_model = self.model
        self.model = parent_model
        self.request = request
        self.object = instance
    

class GenericInlineFormSetMixin(BaseGenericInlineFormSetMixin, InlineFormSetMixin):
    pass


class BaseGenericInlineFormSetView(GenericInlineFormSetMixin, BaseInlineFormSetView):
    pass


class GenericInlineFormSetView(BaseGenericInlineFormSetView, InlineFormSetView):
    pass


from django.db.models import Q
import datetime
import operator


class SearchableListMixin(object):
    """
    Filter queryset like a django admin search_fields does, but with little more intellegence:
    if self.search_split is set to True (by default) it will split query to words (by whatespace)
    Also tries to convert each word to date with self.search_date_formats and then search each word in separate field
    e.g. with query 'foo bar' you can find object with obj.field1='foo' and obj.field2='bar'

    This class is designed to be used with django.generic.ListView

    You could specify query by overriding get_search_query method
    by default this method will try to get 'q' key from request.GET (this can be disabled with search_use_q=False)
    """
    search_fields = ['id']
    search_date_fields = None
    search_date_formats = ['%d.%m.%y', '%d.%m.%Y']
    search_split = True
    search_use_q = True

    def get_words(self, query):
        if self.search_split:
            return query.split()
        return query

    def try_convert_to_date(self, word):
        """
        Tries to convert word to date(datetime) using search_date_formats
        Return None if word fits no one format
        """
        for frm in self.search_date_formats:
            try:
                return datetime.datetime.strptime(word, frm).date()
            except ValueError:
                pass
        return None

    def get_search_query(self):
        """
        Get query from request.GET 'q' parameter when search_use_q is set to True
        Override this method to provide your own query to search
        """
        return self.search_use_q and self.request.GET.get('q', '') or None

    def get_queryset(self):
        qs = super(SearchableListMixin, self).get_queryset()
        query = self.get_search_query()
        if query:
            w_qs = []
            for word in self.get_words(query):
                filters = [Q(**{'%s__icontains' % field_name: word}) for field_name in self.search_fields]
                if self.search_date_fields:
                    dt = self.try_convert_to_date(word)
                    if dt:
                        filters.extend([Q(**{field_name: dt}) for field_name in self.search_date_fields])
                w_qs.append(reduce(operator.or_, filters))
            qs = qs.filter(reduce(operator.and_, w_qs))
        return qs.distinct()


from extra_views.formsets import FormSetView, ModelFormSetView, InlineFormSetView
from extra_views.advanced import CreateWithInlinesView, UpdateWithInlinesView, InlineFormSet, NamedFormsetsMixin
from extra_views.dates import CalendarMonthArchiveView
from extra_views.search import SearchableListMixin
from extra_views.sorting import SortableListMixin


from setuptools import setup

setup(
    name='django-extra-views-ng',
    version='0.3.1',
    url='https://github.com/hovel/django-extra-views',
    description="Extra class-based views for Django",
    long_description=open('README.rst', 'r').read(),
    license="MIT",
    author="Pavel Zhukov",
    author_email="gelios@gmail.com",
    packages=['extra_views'],
    package_dir={'': '.'},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',   
        'Programming Language :: Python']
)
